#ifndef QLEARNINGMATRIX_H
#define QLEARNINGMATRIX_H
//------------------------------------------------------------------------
//
//  Name:   QLearningMatrix.h
//
//  Desc:   A class defining a Q-Learning Matrix.
//
//  Author: Joao Fabro 2014 (fabro@utfpr.edu.br)
//
//------------------------------------------------------------------------
#include <iostream>
using std::cout;
using std::ifstream;
using std::ofstream;
using std::endl;
using std::string;
#include <fstream>
//#include <random>
#include <cassert>
#include <cstdlib>
//using namespace std;
#include<string.h>


#define MAXCOLUMNS 2
#define MAXLINES 50

//#include <rcsc/player/player_agent.h>

//JAF2014: PersistentMatrix is used for the Hill Climbing part
class PersistentMatrixWithState
{
private:
    int id;
    string name;
    int num_columns;
    int num_lines;
    float pMatrix[MAXLINES][MAXCOLUMNS];
    int state;//just used to know which line to "read" next

public:

    PersistentMatrixWithState(){id=0;name="";num_columns=MAXCOLUMNS;num_lines=MAXLINES;}
    PersistentMatrixWithState(int _nc, int _nl){PersistentMatrixWithState();num_columns=_nc;num_lines=_nl;}

    void setId(int _id){id=_id;}

    bool setValue(int _lin, int _col, float _val){
        if ((_lin>=num_lines)||(_col>=num_columns)) return(false);
        pMatrix[_lin][_col]=_val;return true;
    }

    void setState(int _state){state=_state;}//just used to know which line to "read" next
    int getState(){return state;}//just used to know which line to "read" next
    void resetState(){state=0;}//just used to know which line to "read" next

    void setName(string _name){name=_name;}
    void setNumLines(int _nl){num_lines=_nl;}
    void setNumColumns(int _nc){num_columns=_nc;}
    float getValue(int _line, int _col){return pMatrix[_line][_col];}
    void showMatrix()
    {
        for (int i = 0; i < num_lines; i++) {
            for (int j = 0; j < (num_columns); j++) {
                cout << pMatrix[i][j] << "\t";
            }
            cout << endl;
        }
    }
    bool loadControlInformation(){
        string control_suffix(".control");
        string base_name(name);
        base_name.append(control_suffix);
        const char * c = base_name.c_str();

        ifstream iMatrix (c);
        if(iMatrix  && !iMatrix.eof()){
            iMatrix>>id;
            iMatrix>>state;
            iMatrix>>num_lines;
            iMatrix>>num_columns;
        }
        cout << "LoadControlInformation for matrix " << name << " into file " << base_name <<
             " : id=" << id << " state="<< state << " num_lines=" << num_lines << " num_columns=" << num_columns << endl;
        iMatrix.close();
        return(true);
    }

    bool saveControlInformation(){
        char space = '\t';
        string control_suffix(".control");
        string base_name(name);
        base_name.append(control_suffix);
        const char * c = base_name.c_str();
        cout << "SaveControlInformation for matrix " << name << " into file " << base_name <<
             " : id=" << id << " state="<< state << " num_lines=" << num_lines << " num_columns=" << num_columns << endl;

        ofstream oMatrixQS(c);
        oMatrixQS<<id<<space;
        oMatrixQS<<state<<space;
        oMatrixQS<<num_lines<<space;;
        oMatrixQS<<num_columns<<space;;
        oMatrixQS<<endl;
        oMatrixQS.close();
        return(true);
    }

    bool loadMatrix(){
        const char * c = name.c_str();
        loadControlInformation();
        ifstream iMatrix (c);
        while ( iMatrix  && !iMatrix.eof()) {
              for (int i = 0; i < num_lines; i++) {
                  for (int j = 0; j < num_columns; j++) {
                      iMatrix>>pMatrix[i][j];
                  }
              }
         }

         iMatrix.close();
         return(true);
    }

    bool saveMatrix(){
        const char * c = name.c_str();
        ofstream oMatrixQS(c);
        char space = '\t';
        for (int i = 0; i < num_lines; i++) {
            for (int j = 0; j < num_columns; j++) {
                oMatrixQS<< pMatrix[i][j]<<space;
            }
            oMatrixQS<<endl;
        }
        oMatrixQS.close();
        saveControlInformation();
        return(true);
    }

};


#define MAXSTATES 20
#define MAXACTIONS 20

class QLearningMatrix
{
private:

  int id;
  string name;
  int num_of_states;
  int num_of_actions;
  int matrixQ[MAXSTATES][MAXACTIONS];
  int matrixR[MAXSTATES][MAXACTIONS];
  int currentState;
  int previousState;
  int Action;
//  rcsc::PlayerAgent*    agent;
  //location_type         m_Location;
  float alpha;
  float gama;
  float E;

public:

  QLearningMatrix(){id=0;name="";alpha=0.5;gama=0.8;E=0.2;num_of_states=MAXSTATES;num_of_actions=MAXACTIONS;
                    for(int i=0;i<MAXSTATES;i++) for(int j=0;j<MAXACTIONS;j++) matrixQ[i][j]=0;}
  void setId(int _id){id=_id;}
  void setName(string _name){name=_name;}
  void setalpha(float _alpha){alpha=_alpha;}
  void setgama(float _gama){gama=_gama;}
  void setE(float _E){E=_E;}
  void setnum_of_states(int _num_of_states){num_of_states=_num_of_states;}
  void setnum_of_actions(int _num_of_actions){num_of_actions=_num_of_actions;}
  void setState(int _state){cout << "SP2014:updateInternalState(): QLearningMatrix updated currentState to " << _state << endl;
                            cout.flush();
                            currentState=_state;}
  void setPreviousState(int _state){cout << "SP2014:updateInternalState(): QLearningMatrix updated previousState to " << _state << endl;
                                    cout.flush();
                                    previousState=_state;}
  //rcsc::PlayerAgent* GetAgent() const {return agent;}

  int getId() const{return id;}
  string getName() const{return name;}
  int getcurrentState() const{return currentState;}
  //getCellMatrixQ(int state, int action) returns the value at lin=state, col=action.
  int getCellmatrixQ(int state, int action){return matrixQ[state][action];}
  void change_state_executing_action(int _actionNumber, int _newState)
  {
      Action = _actionNumber;
      cout << "SP2014:updateInternalState(): QLearningMatrix moving from state " << currentState << " to state " << _newState << " by executing action " << _actionNumber << endl;
      cout.flush();
      previousState = currentState;
      currentState = _newState;
  }

  void showMatrix(int matrix[MAXSTATES][MAXACTIONS])
  {
      for (int i = 0; i < num_of_states; i++) {
          for (int j = 0; j < (num_of_actions-2); j++) {
              cout << matrix[i][j] << "\t";
          }
          cout << endl;
      }
  }

  void showMatrixQ(){cout << "Printing Matrix Q :" << name << id << endl;showMatrix(matrixQ);}
  void showMatrixR(){cout << "Printing Matrix R :" << name << id << endl;showMatrix(matrixR);}

  bool loadMatrix(char *matrixname, int matrix[MAXSTATES][MAXACTIONS]){

        //loads matrix
        ifstream iMatrix (matrixname);
        while ( iMatrix  && !iMatrix.eof()) {
            for (int i = 0; i < num_of_states; i++) {
                for (int j = 0; j < num_of_actions; j++) {
                    iMatrix>>matrix[i][j];
                }
            }
        }
        iMatrix.close();
        return(true);
  }

  bool loadMatrixQ(){
      char matrixq[100]="matrixQ";//+name;//+(string)id;
      strcat(matrixq,name.c_str());

      cout << "SP2014: Loading: " << matrixq;
      if(loadMatrix(matrixq,matrixQ))
      {
       cout << "SP2014: Loaded MatrixQ: " << matrixq << " with " << num_of_states << " states" <<
                " and " << num_of_actions << " actions" << endl;cout.flush();
       return(true);
      }
      else
          return(false);
  }

  bool loadMatrixR(){
      char matrixq[100]="matrixR";//+name;//+(string)id;
      strcat(matrixq,name.c_str());

      cout << "SP2014: Loading: " << matrixq;
      if(loadMatrix(matrixq,matrixR))
      {
      cout << "SP2014: Loaded MatrixR: " << matrixq << " with " << num_of_states << " states" <<
                " and " << num_of_actions << " actions" << endl;cout.flush();
      return(true);
      }
      else
          return(false);
  }

  bool saveMatrix(char *matrixname){
      char space = ' ';
      ofstream oMatrixQS(matrixname);

      for (int i = 0; i < num_of_states; i++) {
          for (int j = 0; j < num_of_actions; j++) {
              oMatrixQS<< matrixQ[i][j]<<space;
          }
          oMatrixQS<<endl;
      }
      oMatrixQS.close();
      return(true);
  }

  bool saveMatrixQ(){
      char matrixq[100] = "matrixQ";// + name;// + (string)id;
      strcat(matrixq,name.c_str());
      saveMatrix(matrixq);
      cout << "SP2014: Saved MatrixQ: " << matrixq << " with " << num_of_states << " states" <<
                " and " << num_of_actions << " actions" << endl;cout.flush();
      return(true);
  }

  int getIndexActionforCurrentState()
  {
      int acao;
      float q = (rand()%10) ;

      q = q /10.0;
      cout<<"\n"<<"q : "<<q;


      if (q <= E) { //escolha uma acao aleatoria
       do {
              acao = rand()%num_of_actions;
              //cout<<"\n"<<"acao : "<<acao;
              //cout<<"\n"<<matrizR[estado][acao];
          } while (matrixQ[currentState][acao] == 0);
      }
      else { // encontre acao maxima
          int max1 = -1;

          for ( int i = 0; i < num_of_actions; i++ ) {

              if (matrixQ[currentState][i] != 0) {

                  if (matrixQ[currentState][i] > max1) {
                      max1 = matrixQ[currentState][i];
                      acao = i;
                  }
              }
          }
          cout<<"\n"<<"SP2014: Selected Action: " << acao << " Value:" << max1 << endl;
          cout.flush();
      }
      return(acao);
  }

  bool updateMatrixQ(float _reinforcement)
  {
      int max2;
      float test;
      if (currentState != previousState) {
//         previousState = 0;//JAF2014 HACK to convert graph to tree - and use one 1 line of the matrix.....
//         cout << "updateMatrixQ(reinforcement): previousstate=0" << endl;cout.flush();//JAF2014 HACK
         max2 = 0;
         for (int i = 0; i < num_of_actions; i++ ) {
//             if (matrixQ[0][i] >= max2) {//JAF2014 HACK
//                 max2 = matrixQ[0][i];   //JAF2014 HACK
//             }                           //JAF2014 HACK
              if (matrixQ[currentState][i] >= max2) {
                  max2 = matrixQ[currentState][i];
              }
         }
         //Max 2 is the MAXIMUM value of Q Matrix for the current State!!!

         //if(max2==0) return(false);//Problem with finding the selected action value.....

         cout<<"\n"<<"max2 : "<<max2<< endl;
         cout<<" previousState = " << previousState << " Action = " <<Action << endl;
         cout<<" matrixQ[previousState][Action] = " << matrixQ[previousState][Action] << endl;
         cout<<" _reinforcement = " << _reinforcement << endl;
         cout<<"matrixQ[previousState][Action] = " <<  matrixQ[previousState][Action] << " + alpha=" << alpha << " * (";
         cout<< _reinforcement << " + gama=" << gama << " * max2=" << max2 << " - " << matrixQ[previousState][Action] << " ) " << endl;

         test =  matrixQ[previousState][Action] + alpha *(_reinforcement
                                                          + gama * max2
                                                          -  matrixQ[previousState][Action]);

         matrixQ[previousState][Action] = test;

         cout<<"\n"<<"Value of new Q for State: "<< previousState << " and Action: " << Action << " : " << test << endl;cout.flush();
         return(true);
    } else
         return(false);
 }

  bool updateMatrixQ(float _reinforcement, int max2)
  {
 //    int max2;
      float test;
      if (currentState != previousState) {
         previousState = 0;//JAF2014 HACK to convert graph to tree - and use one 1 line of the matrix.....
//         max2 = 0;
//         for (int i = 0; i < num_of_actions; i++ ) {

//              if (matrixQ[currentState][i] >= max2) {
//                  max2 = matrixQ[currentState][i];
//              }
//         }
//         //Max 2 is the MAXIMUM value of Q Matrix for the current State!!!

//         //if(max2==0) return(false);//Problem with finding the selected action value.....

         cout<<"\n"<<"max2 : "<<max2<< endl;
         cout<<" previousState = " << previousState << " Action = " <<Action << endl;
         cout<<" matrixQ[previousState][Action] = " << matrixQ[previousState][Action] << endl;
         cout<<" _reinforcement = " << _reinforcement << endl;
         cout<<"matrixQ[previousState][Action] = " <<  matrixQ[previousState][Action] << " + alpha=" << alpha << " * (";
         cout<< _reinforcement << " + gama=" << gama << " * max2=" << max2 << " - " << matrixQ[previousState][Action] << " ) " << endl;

         test =  matrixQ[previousState][Action] + alpha *(_reinforcement
                                                          + gama * max2
                                                          -  matrixQ[previousState][Action]);

         matrixQ[previousState][Action] = test;

         cout<<"\n"<<"Value of new Q for State: "<< previousState << " and Action: " << Action << " : " << test << endl;cout.flush();
         return(true);
    } else
         return(false);
 }

  bool updateMatrixQ()
  {
      int max2;
      float test;
      if (currentState != previousState) {
         max2 = 0;
         for (int i = 0; i < num_of_actions; i++ ) {

              if (matrixQ[currentState][i] >= max2) {
                  max2 = matrixQ[currentState][i];
              }
         }
         //Max 2 is the MAXIMUM value of Q Matrix for the current State!!!

         //if(max2==0) return(false);//Problem with finding the selected action value.....

         cout<<"\n"<<"max2 : "<<max2<< endl;
         cout<<" previousState = " << previousState << " Action = " <<Action << endl;
         cout<<" matrixQ[previousState][Action] = " << matrixQ[previousState][Action] << endl;
         cout<<" reinforcement = " << matrixR[previousState][Action] << endl;
         cout<<"matrixQ[previousState][Action] = " <<  matrixQ[previousState][Action] << " + alpha=" << alpha << " * (";
         cout<< matrixR[previousState][Action] << " + gama=" << gama << " * max2=" << max2 << " - " << matrixQ[previousState][Action] << " ) " << endl;

         test =  matrixQ[previousState][Action] + alpha *(matrixR[previousState][Action]
                                                          + gama * max2
                                                          -  matrixQ[previousState][Action]);

         matrixQ[previousState][Action] = test;

         cout<<"\n"<<"Value of new Q for State: "<< previousState << " and Action: " << Action << " : " << test << endl;cout.flush();
         return(true);
    } else
         return(false);
 }
};

#endif // QLEARNINGMATRIX_H
